# This file is Copyright 2024 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import logging
import struct
from typing import List, Set

from volatility3.framework import exceptions, constants
from volatility3.framework import renderers
from volatility3.framework.renderers import format_hints
from volatility3.framework.configuration import requirements
from volatility3.framework.interfaces import plugins
from volatility3.framework.symbols import linux
from volatility3.framework import symbols
from volatility3.plugins.linux import lsof, pslist, sockstat
from volatility3.framework.layers import scanners
from volatility3.framework.symbols.linux import network

vollog = logging.getLogger(__name__)


class Sockscan(plugins.PluginInterface):
    """Scans for network connections found in memory layer."""

    _required_framework_version = (2, 6, 0)
    _version = (1, 0, 0)

    @classmethod
    def get_requirements(cls):
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Linux kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="SockHandlers", component=sockstat.SockHandlers, version=(4, 0, 0)
            ),
            requirements.VersionRequirement(
                name="lsof", component=lsof.Lsof, version=(2, 0, 0)
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(4, 1, 0)
            ),
            requirements.VersionRequirement(
                name="linuxutils", component=linux.LinuxUtilities, version=(2, 1, 0)
            ),
            requirements.VersionRequirement(
                name="linux_net", component=network.NetSymbols, version=(1, 0, 0)
            ),
            requirements.VersionRequirement(
                name="multi_string_scanner",
                component=scanners.MultiStringScanner,
                version=(1, 0, 0),
            ),
        ]

    def _canonicalize_symbol_addrs(
        self, kernel_module_name: str, symbol_names: List[str]
    ) -> Set[bytes]:
        """Takes a list of symbol names and converts the address of each to the bytes
        as they would appear in memory so that they can be scanned for.

        Symbols that cannot be found are ignored and not included in the results.

        Args:
            kernel_module_name: The name of the kernel module on which to operate
            symbol_names: A list of symbol names to be looked up

        Returns:
            A set of bytes which are the packed addresses.
        """
        # get vmlinux module from context in order to build objects and read symbols
        vmlinux = self.context.modules[kernel_module_name]

        # get kernel layer from context so that it's dependencies can be found, and therefore scanned.
        # kernel layer will be virtual and built ontop of a physical layer.
        kernel_layer = self.context.layers[vmlinux.layer_name]

        # detmine if kernel is 64bit or not. The plugin scans for pointers and these need to formated
        # to the correct size so that they can be accurately located in the physical layer.
        if symbols.symbol_table_is_64bit(self.context, vmlinux.symbol_table_name):
            pack_format = "Q"  # 64 bit
        else:
            pack_format = "I"  # 32 bit

        packed_needles = set()
        for symbol_name in symbol_names:
            try:
                needle_addr = vmlinux.object_from_symbol(symbol_name).vol.offset
            except exceptions.SymbolError:
                vollog.log(
                    constants.LOGLEVEL_VVVV,
                    f"Unable to find symbol {symbol_name} this will not be scanned for.",
                )
                continue
            # use canonicalize to set the appropriate sign extension for the addr
            addr = kernel_layer.canonicalize(needle_addr)
            packed_addr = struct.pack(pack_format, addr)
            packed_needles.add(packed_addr)
            vollog.log(
                constants.LOGLEVEL_VVVV,
                f"Will scan for {symbol_name} using the bytes: {packed_addr.hex()}",
            )

        # make a warning if no symbols at all could be resolved.
        if not packed_needles:
            vollog.warning(
                "_canonicalize_symbol_addrs was unable to resolve any symbols, use -vvvv for more information."
            )

        return packed_needles

    def _find_memory_layer_name(self, kernel_module_name: str):
        """Find the memory layer below the kernel. Only returns a single layer,
        and will warn the user if multiple layers are found.

        Args:
            kernel_module_name: The name of the kernel module on which to operate.

        Returns:
            memory_layer_name: The name of the layer below the kernel to be scanned.
        """

        # get vmlinux module from context in order to build objects and read symbols
        vmlinux = self.context.modules[kernel_module_name]

        # get kernel layer from context so that it's dependencies can be found, and therefore scanned.
        # kernel layer will be virtual and built ontop of a physical layer.
        kernel_layer = self.context.layers[vmlinux.layer_name]

        # TODO: Update plugin to support multiple dependencies. e.g. a memory layer and swap file.
        # This is a shared problem with psscan and having a generic solution would be useful.

        # Find the memory layer to scan, and provide warnings if more than one is located.
        if len(kernel_layer.dependencies) > 1:
            vollog.warning(
                f"Kernel layer depends on multiple layers however only {kernel_layer.dependencies[0]} will be scanned by this plugin."
            )
        elif len(kernel_layer.dependencies) == 0:
            vollog.error(
                "Kernel layer has no dependencies, meaning there is no memory layer for this plugin to scan."
            )
            raise exceptions.LayerException(
                vmlinux.layer_name, f"Layer {vmlinux.layer_name} has no dependencies"
            )

        memory_layer_name = kernel_layer.dependencies[0]

        return memory_layer_name

    def _find_file_ops_needles(self, kernel_module_name: str):
        """Retrieves socket file symbols and the offset to the 'f_op' pointer.

        Args:
            kernel_module_name (str): The name of the kernel module to search.

        Returns:
            Tuple[List[int], int]: A list of file symbol addresses and,
            the offset to the 'f_op' pointer.
        """

        # get vmlinux module from context in order to read symbols
        vmlinux = self.context.modules[kernel_module_name]

        file_ops_symbol_names = [
            "socket_file_ops",
            "sockfs_dentry_operations",
        ]
        file_ops_needles = self._canonicalize_symbol_addrs(
            kernel_module_name, file_ops_symbol_names
        )
        # get file struct to find the offset to the f_op pointer
        # this is so that the file object can be created at the correct offset,
        # the results of the scanner will be for the f_op member within the file
        f_op_offset = vmlinux.get_type("file").relative_child_offset("f_op")

        return (file_ops_needles, f_op_offset)

    def _find_sk_destruct_needles(self, kernel_module_name: str):
        # get vmlinux module from context in order to read symbols
        vmlinux = self.context.modules[kernel_module_name]

        socket_destructor_symbol_names = [
            "sock_def_destruct",
            "packet_sock_destruct",
            "unix_sock_destructor",
            "netlink_sock_destruct",
            "inet_sock_destruct",
        ]
        socket_destructor_needles = self._canonicalize_symbol_addrs(
            kernel_module_name, socket_destructor_symbol_names
        )
        # get sock struct to find the offset to the sk_destruct pointer
        # this is so that the sock object can be created at the correct offset,
        # the results of the scanner will be for the sk_destruct member within the scock
        sk_destruct_offset = vmlinux.get_type("sock").relative_child_offset(
            "sk_destruct"
        )
        return (socket_destructor_needles, sk_destruct_offset)

    def _walk_file_ops_needles(
        self,
        kernel_module_name: str,
        physical_memory_layer_name: str,
        needle_addr: int,
        f_op_offset: int,
    ):
        """
        This method attempts to walk from the f_op member of files to the
        corresponding socket. If sucessful the socket object is created on the
        memory layer and returned.

        Args:
            kernel_module_name (str): The name of the kernel module from which,
            to retrieve the file operations.
            physical_memory_layer_name (str): The name of the physical memory layer that was scanned
            needle_addr: The address of the needle that was found during the scanning
            f_op_offset: The offset to the f_op member of the file type

        Returns:
            psock: The sock object that was built on the memory layer
        """

        vmlinux = self.context.modules[kernel_module_name]
        try:
            # create file in the memory_layer, the native layer matches the
            # kernel so that pointers can be followed
            sock_physical_addr = needle_addr - f_op_offset
            pfile = self.context.object(
                vmlinux.symbol_table_name + constants.BANG + "file",
                offset=sock_physical_addr,
                layer_name=physical_memory_layer_name,
                native_layer_name=vmlinux.layer_name,
            )
            dentry = pfile.get_dentry()
            if not dentry:
                vollog.log(
                    constants.LOGLEVEL_VVVV,
                    f"Skipping file at {hex(needle_addr)} as unable to locate dentry",
                )
                return None

            d_inode = dentry.d_inode
            if not d_inode:
                vollog.log(
                    constants.LOGLEVEL_VVVV,
                    f"Skipping file at {hex(needle_addr)} as unable to locate inode for dentry",
                )
                return None

            socket_alloc = linux.LinuxUtilities.container_of(
                d_inode, "socket_alloc", "vfs_inode", vmlinux
            )
            socket = socket_alloc.socket
            if not (socket and socket.sk):
                vollog.log(
                    constants.LOGLEVEL_VVVV,
                    f"Skipping file at {hex(needle_addr)} as socket created by LinuxUtilities.container_of is invalid",
                )
                return None

            # sucessfully trversed from file to sock, this will exist in the
            # kernel layer, and need to be translated to the memory layer.
            vsock = socket.sk.dereference()

            # get virtual offset
            virtual_sock_offset = vsock.vol.offset

            # translate this offset to physical
            native_layer = self.context.layers[vmlinux.layer_name]
            physical_sock_offset, _physical_layer_name = native_layer.translate(
                virtual_sock_offset
            )

            # build sock on the memory_layer using the physical_sock_offset
            psock = self.context.object(
                vmlinux.symbol_table_name + constants.BANG + "sock",
                offset=physical_sock_offset,
                layer_name=physical_memory_layer_name,
                native_layer_name=vmlinux.layer_name,
            )

            return psock

        except exceptions.InvalidAddressException as error:
            vollog.log(
                constants.LOGLEVEL_VVVV,
                f"Unable to follow file at {hex(needle_addr)} to socket due to invalid address: {error}",
            )
        return None

    def _extract_sock_fields(self, psock, sock_handler):
        try:
            sock_physical_addr = psock.vol.offset
            sock_type = psock.get_type()

            family = psock.get_family()
            # remove results with no family
            if family is None:
                vollog.log(
                    constants.LOGLEVEL_VVVV,
                    f"Skipping socket at {hex(sock_physical_addr)} as unable to determin family.",
                )
                return None

            # TODO: invesitgate options for more invalid address handling in proccess_sock
            # and the later formatting of it's results.
            sock_fields = sock_handler.process_sock(psock)
            # if no sock_fields we're able to be extracted then skip this result.
            if not sock_fields:
                vollog.log(
                    constants.LOGLEVEL_VVVV,
                    f"Skipping socket at {hex(sock_physical_addr)} as unable to process with SockHandlers.",
                )
                return None

            sock, sock_stat, extended = sock_fields
            src, src_port, dst, dst_port, state = sock_stat
            protocol = sock.get_protocol()

            # format results
            src = renderers.NotAvailableValue() if src is None else str(src)
            src_port = (
                renderers.NotAvailableValue() if src_port is None else str(src_port)
            )
            dst = renderers.NotAvailableValue() if dst is None else str(dst)
            dst_port = (
                renderers.NotAvailableValue() if dst_port is None else str(dst_port)
            )
            state = renderers.NotAvailableValue() if state is None else str(state)
            protocol = (
                renderers.NotAvailableValue() if protocol is None else str(protocol)
            )
            # extended attributes is a dict, so this is formated to string show each
            # key and value pair, seperated with a comma.
            socket_filter_str = (
                ",".join(f"{k}={v}" for k, v in extended.items())
                if extended
                else renderers.NotAvailableValue()
            )

            # remove empty results
            if (src == "0.0.0.0" or isinstance(src, renderers.NotAvailableValue)) and (
                dst == "0.0.0.0" or isinstance(src, renderers.NotAvailableValue)
            ):
                if state == "UNCONNECTED":
                    return None
                elif src_port == "0" and dst_port == "0":
                    return None
            return (
                format_hints.Hex(sock_physical_addr),
                family,
                sock_type,
                protocol,
                src,
                src_port,
                dst,
                dst_port,
                state,
                socket_filter_str,
            )

        except exceptions.InvalidAddressException as error:
            vollog.log(
                constants.LOGLEVEL_VVVV,
                f"Unable create results for socket at {hex(sock_physical_addr)} due to invalid address: {error}",
            )
        return None

    def _generator(self, kernel_module_name: str):
        """Scans for sockets. Each row represents a kernel socket.

        Args:
            kernel_module_name: The name of the kernel module on which to operate

        Yields:
            addr: Physical offset
            family: Socket family string (AF_UNIX, AF_INET, etc)
            sock_type: Socket type string (STREAM, DGRAM, etc)
            protocol: Protocol string (UDP, TCP, etc)
            source addr: Source address string
            source port: Source port string (not all of them are int)
            destination addr: Destination address string
            destination port: Destination port (not all of them are int)
            state: State strings (LISTEN, CONNECTED, etc)
        """

        # get vmlinux module from context in order to build objects and read symbols
        vmlinux = self.context.modules[kernel_module_name]

        # get the memory layer that is to be scanned.
        memory_layer_name = self._find_memory_layer_name(kernel_module_name)
        memory_layer = self.context.layers[memory_layer_name]

        # use the init process to build a sock handler
        # TODO: look into options so that sockstat.SockHandlers so that process_sock can
        # be used without a task object.
        init_task = vmlinux.object_from_symbol(symbol_name="init_task")
        sock_handler = sockstat.SockHandlers(
            self.context, kernel_module_name, init_task
        )

        # get progress_callback in order to use this in the scanners.
        # TODO: perhaps add more detail to progress, showing method in progress and number of hits found
        progress_callback = self._progress_callback

        # Method 1 - find sockets by file operations, then follow pointers to sockets
        file_ops_needles, f_op_offset = self._find_file_ops_needles(kernel_module_name)

        # Method 2 - find sockets by socket destructor directly inside sock objects
        socket_destructor_needles, sk_destruct_offset = self._find_sk_destruct_needles(
            kernel_module_name
        )

        # TODO Method 3 - find sock by sk_error_report symbols
        # sk_error_report_symbol_names = ['sock_def_error_report', 'inet_sk_rebuild_header', 'inet_listen']
        # this would be similar to Method 2, but using a different pointer within sock.

        # add a set of seen addresses to stop possible duplication of results.
        seen_sock_physical_addr = set()

        # Using the calculated needles, scan the memory layer and attempt to parse the sockets located.
        for needle_addr, match in memory_layer.scan(
            self.context,
            scanners.MultiStringScanner(socket_destructor_needles | file_ops_needles),
            progress_callback,
        ):
            psock = None
            sock_physical_addr = None

            # if match is from socket_destructor_needles simply calculate the offset to the sock
            if match in socket_destructor_needles:
                sock_physical_addr = needle_addr - sk_destruct_offset
                psock = self.context.object(
                    vmlinux.symbol_table_name + constants.BANG + "sock",
                    offset=sock_physical_addr,
                    layer_name=memory_layer_name,
                    native_layer_name=vmlinux.layer_name,
                )

            # if match is from file_ops_needles attempt to walk from file object to the sock
            if match in file_ops_needles:
                psock = self._walk_file_ops_needles(
                    kernel_module_name, memory_layer_name, needle_addr, f_op_offset
                )

            if psock is not None and sock_physical_addr not in seen_sock_physical_addr:
                seen_sock_physical_addr.add(sock_physical_addr)

                fields = self._extract_sock_fields(psock, sock_handler)
                if fields:
                    yield (0, fields)

    def run(self):

        tree_grid_args = [
            ("Sock Offset", format_hints.Hex),
            ("Family", str),
            ("Type", str),
            ("Proto", str),
            ("Source Addr", str),
            ("Source Port", str),
            ("Destination Addr", str),
            ("Destination Port", str),
            ("State", str),
            ("Filter", str),
        ]

        return renderers.TreeGrid(
            tree_grid_args,
            self._generator(self.config["kernel"]),
        )



# This file is Copyright 2025 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
import logging
from typing import List, Tuple, Generator, Optional

from volatility3.framework import renderers, interfaces, constants, exceptions
from volatility3.framework.renderers import format_hints
from volatility3.framework.configuration import requirements
from volatility3.framework.interfaces import plugins
from volatility3.framework.objects import utility
from volatility3.plugins.linux import pslist

vollog = logging.getLogger(__name__)


class PerfEvents(plugins.PluginInterface):
    """Lists performance events for each process."""

    _required_framework_version = (2, 0, 0)
    _version = (1, 0, 1)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Linux kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(4, 0, 0)
            ),
        ]

    @classmethod
    def list_perf_events(
        cls, context, vmlinux_module_name: str
    ) -> Generator[
        Tuple[
            interfaces.objects.ObjectInterface,
            interfaces.objects.ObjectInterface,
            Optional[str],
            Optional[str],
            Optional[str],
            Optional[int],
        ],
        None,
        None,
    ]:
        """
        Walks the `perf_event_list` of each `task_struct` and reports valid event structures found
        This plugin is one of several to detect eBPF based malware

        Args:
            context:
            vmlinux_module_name:

        Returns:
            A tuple of the task struct, performance event object, event name, program name, full name, and program address
        """
        vmlinux = context.modules[vmlinux_module_name]

        if not vmlinux.has_type("perf_event") or not vmlinux.get_type(
            "perf_event"
        ).has_member("owner_entry"):
            vollog.warning(
                "This kernel does not have performance events enabled (CONFIG_PERF_EVENTS). Cannot proceed."
            )
            return

        for task in pslist.PsList.list_tasks(
            context, vmlinux_module_name, include_threads=True
        ):
            # walk the list of perf_event entries for this process
            for event in task.perf_event_list.to_list(
                vmlinux.symbol_table_name + constants.BANG + "perf_event", "owner_entry"
            ):
                # if the names are smeared then bail
                try:
                    event_name = utility.pointer_to_string(event.pmu.name, count=64)
                    try:
                        full_name = utility.array_to_string(
                            event.prog.aux.ksym.name, count=512
                        )
                    except AttributeError:
                        full_name = None

                    program_name = utility.array_to_string(event.prog.aux.name)
                except exceptions.InvalidAddressException:
                    continue

                # if the kernel has the prog member then ensure it is not 0
                if hasattr(event, "prog"):
                    program_address = event.prog
                    if program_address == 0:
                        continue

                else:
                    program_address = None

                yield task, event_name, program_name, full_name, program_address

    def _generator(self):
        for (
            task,
            event_name,
            program_name,
            full_name,
            program_address,
        ) in self.list_perf_events(self.context, self.config["kernel"]):
            task_name = utility.array_to_string(task.comm)

            # We at least need one useful string...
            if event_name is None and program_name is None and full_name is None:
                continue

            if program_address is not None:
                program_address = format_hints.Hex(program_address)
            else:
                program_address = renderers.NotAvailableValue()

            yield (
                0,
                (
                    task.pid,
                    task_name,
                    event_name or renderers.NotAvailableValue(),
                    program_name or renderers.NotAvailableValue(),
                    full_name or renderers.NotAvailableValue(),
                    program_address,
                ),
            )

    def run(self) -> renderers.TreeGrid:
        return renderers.TreeGrid(
            [
                ("PID", int),
                ("Process", str),
                ("Event", str),
                ("Short Program Name", str),
                ("Full Name", str),
                ("Address", format_hints.Hex),
            ],
            self._generator(),
        )



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import logging
import struct
from datetime import datetime
from typing import Dict, Optional, Tuple, Union

from volatility3.framework import constants, exceptions, interfaces, objects, renderers
from volatility3.framework.objects.utility import address_to_string
from volatility3.framework.symbols.windows.extensions import conversion

vollog = logging.getLogger(__name__)


class SHIM_CACHE_ENTRY(objects.StructType):
    """Class for abstracting variations in the shimcache LRU list entry structure"""

    def __init__(
        self,
        context: interfaces.context.ContextInterface,
        type_name: str,
        object_info: interfaces.objects.ObjectInformation,
        size: int,
        members: Dict[str, Tuple[int, interfaces.objects.Template]],
    ) -> None:
        super().__init__(context, type_name, object_info, size, members)
        self._exec_flag = None
        self._file_path = None
        self._file_size = None
        self._last_modified = None
        self._last_updated = None

    @property
    def exec_flag(self) -> Union[bool, interfaces.renderers.BaseAbsentValue]:
        """Checks if InsertFlags fields has been bitwise OR'd with a value of 2.
        This behavior was observed when processes are created by CSRSS."""
        if self._exec_flag is not None:
            return self._exec_flag

        try:
            if hasattr(self, "ListEntryDetail") and hasattr(
                self.ListEntryDetail, "InsertFlags"
            ):
                self._exec_flag = self.ListEntryDetail.InsertFlags & 0x2 == 2

            elif hasattr(self, "InsertFlags"):
                self._exec_flag = self.InsertFlags & 0x2 == 2

            elif hasattr(self, "ListEntryDetail") and hasattr(
                self.ListEntryDetail, "BlobBuffer"
            ):
                blob_offset = self.ListEntryDetail.BlobBuffer
                blob_size = self.ListEntryDetail.BlobSize

                if not self._context.layers[self.vol.native_layer_name].is_valid(
                    blob_offset, blob_size
                ):
                    self._exec_flag = renderers.UnreadableValue()
                    return self._exec_flag

                raw_flag = self._context.layers[self.vol.native_layer_name].read(
                    blob_offset, blob_size
                )
                if not raw_flag:
                    self._exec_flag = renderers.UnparsableValue()
                    return self._exec_flag

                try:
                    self._exec_flag = bool(struct.unpack("<I", raw_flag)[0])
                except struct.error:
                    self._exec_flag = renderers.UnparsableValue()

        except exceptions.InvalidAddressException:
            vollog.debug(
                "Failed to read SHIMCACHE_ENTRY exec flag due to invalid address exception"
            )
            self._exec_flag = renderers.UnreadableValue()

        else:
            # Always set to true for XP/2K3
            self._exec_flag = renderers.NotApplicableValue()

        return self._exec_flag

    @property
    def file_size(self) -> Union[int, interfaces.renderers.BaseAbsentValue]:
        if self._file_size is not None:
            return self._file_size
        try:
            self._file_size = self.FileSize
            if self._file_size < 0:
                self._file_size = 0

        except AttributeError:
            self._file_size = renderers.NotApplicableValue()
        except exceptions.InvalidAddressException:
            self._file_size = renderers.UnreadableValue()

        return self._file_size

    @property
    def last_modified(self) -> Union[datetime, interfaces.renderers.BaseAbsentValue]:
        if self._last_modified is not None:
            return self._last_modified
        try:
            self._last_modified = conversion.wintime_to_datetime(
                self.ListEntryDetail.LastModified.QuadPart
            )
        except AttributeError:
            self._last_modified = conversion.wintime_to_datetime(
                self.LastModified.QuadPart
            )
        except exceptions.InvalidAddressException:
            self._last_modified = renderers.UnreadableValue()

        return self._last_modified

    @property
    def last_update(self) -> Union[datetime, interfaces.renderers.BaseAbsentValue]:
        if self._last_updated is not None:
            return self._last_updated

        try:
            self._last_updated = conversion.wintime_to_datetime(
                self.LastUpdate.QuadPart
            )
        except AttributeError:
            self._last_updated = renderers.NotApplicableValue()
        except exceptions.InvalidAddressException:
            self._last_updated = renderers.UnreadableValue()

        return self._last_updated

    @property
    def file_path(self) -> Union[str, interfaces.renderers.BaseAbsentValue]:
        if self._file_path is not None:
            return self._file_path

        if not hasattr(self.Path, "Buffer"):
            return address_to_string(
                self._context,
                self.Path.vol.layer_name,
                self.Path.vol.offset,
                self.Path.vol.count,
                encoding="utf-16le",
            )

        try:
            file_path_raw = (
                self._context.layers[self.vol.native_layer_name].read(
                    self.Path.Buffer, self.Path.Length
                )
                or b""
            )
            self._file_path = file_path_raw.decode("utf-16", errors="replace")
        except exceptions.InvalidAddressException:
            self._file_path = renderers.UnreadableValue()

        return self._file_path

    def is_valid(self) -> bool:
        """Shim cache validation is limited to ensuring that a subset of the
        pointers in the LIST_ENTRY field are valid (similar to validation of
        ERESOURCE)"""

        # shim entries on Windows XP do not have list entry attributes; in this case,
        # perform a different set of validations
        try:
            if not hasattr(self, "ListEntry"):
                return bool(self.last_modified and self.last_update and self.file_size)

            # on some platforms ListEntry.Blink is null, so this cannot be validated
            if (
                self.ListEntry.Flink != 0
                and (
                    self.ListEntry.Blink.dereference()
                    != self.ListEntry.Flink.dereference()
                )
                and (
                    self.ListEntry.Flink.Blink
                    == self.ListEntry.Flink.Blink.dereference().vol.offset
                )
            ):
                return True
            else:
                return False
        except exceptions.InvalidAddressException:
            return False


class SHIM_CACHE_HANDLE(objects.StructType):
    def __init__(
        self,
        context: interfaces.context.ContextInterface,
        type_name: str,
        object_info: interfaces.objects.ObjectInformation,
        size: int,
        members: Dict[str, Tuple[int, interfaces.objects.Template]],
    ) -> None:
        super().__init__(context, type_name, object_info, size, members)

    @property
    def head(self) -> Optional[SHIM_CACHE_ENTRY]:
        try:
            if not self.eresource.is_valid():
                return None
        except exceptions.InvalidAddressException:
            return None

        rtl_avl_table = self._context.object(
            self.get_symbol_table_name() + constants.BANG + "_RTL_AVL_TABLE",
            self.vol.layer_name,
            self.rtl_avl_table,
            self.vol.native_layer_name,
        )

        if not self._context.layers[self.vol.layer_name].is_valid(
            self.rtl_avl_table.vol.offset
        ):
            return None

        offset_head = rtl_avl_table.vol.offset + rtl_avl_table.vol.size

        head = self._context.object(
            self.get_symbol_table_name() + constants.BANG + "SHIM_CACHE_ENTRY",
            self.vol.layer_name,
            offset_head,
        )

        if not head.is_valid():
            return None

        return head

    def is_valid(self, avl_section_start: int, avl_section_end: int) -> bool:
        if self.vol.offset == 0:
            return False

        vollog.debug(f"Checking SHIM_CACHE_HANDLE validity @ {hex(self.vol.offset)}")

        if not (
            self._context.layers[self.vol.layer_name].is_valid(self.vol.offset)
            and self.eresource.is_valid()
            and self.rtl_avl_table.is_valid(avl_section_start, avl_section_end)
            and self.head
        ):
            return False

        return self.head.is_valid()


class RTL_AVL_TABLE(objects.StructType):
    def is_valid(self, page_start: int, page_end: int) -> bool:
        try:
            if self.BalancedRoot.Parent != self.BalancedRoot.vol.offset:
                vollog.debug(
                    f"RTL_AVL_TABLE @ {self.vol.offset} Invalid: Failed BalancedRoot parent equality check"
                )
                return False

            elif self.AllocateRoutine < page_start or self.AllocateRoutine > page_end:
                vollog.debug(
                    f"RTL_AVL_TABLE @ {self.vol.offset} Invalid: Failed AllocateRoutine range check"
                )
                return False

            elif self.CompareRoutine < page_start or self.CompareRoutine > page_end:
                vollog.debug(
                    f"RTL_AVL_TABLE @ {self.vol.offset} Invalid: Failed CompareRoutine range check"
                )
                return False

            elif (
                (self.AllocateRoutine.vol.offset == self.CompareRoutine.vol.offset)
                or (self.AllocateRoutine.vol.offset == self.FreeRoutine.vol.offset)
                or (self.CompareRoutine.vol.offset == self.FreeRoutine.vol.offset)
            ):
                vollog.debug(
                    f"RTL_AVL_TABLE @ {self.vol.offset} Invalid: Failed (Compare|Allocate|Free)Routine uniqueness check"
                )
                return False

            return True
        except exceptions.InvalidAddressException:
            return False


class_types = {
    "SHIM_CACHE_HANDLE": SHIM_CACHE_HANDLE,
    "SHIM_CACHE_ENTRY": SHIM_CACHE_ENTRY,
    "_RTL_AVL_TABLE": RTL_AVL_TABLE,
}



# This file is Copyright 2024 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
import logging
from typing import Generator, Iterable, List, Set, Tuple

from volatility3.framework import (
    constants,
    deprecation,
    exceptions,
    interfaces,
    renderers,
)
from volatility3.framework.configuration import requirements
from volatility3.framework.interfaces import plugins
from volatility3.framework.symbols.linux import extensions
from volatility3.framework.symbols.linux.utilities import (
    modules as linux_utilities_modules,
)

vollog = logging.getLogger(__name__)


class Hidden_modules(plugins.PluginInterface):
    """Carves memory to find hidden kernel modules"""

    _required_framework_version = (2, 25, 0)
    _version = (3, 0, 3)

    @classmethod
    def find_hidden_modules(
        cls, context, vmlinux_module_name: str
    ) -> Generator[extensions.module, None, None]:
        if context.symbol_space.verify_table_versions(
            "dwarf2json", lambda version, _: (not version) or version < (0, 8, 0)
        ):
            raise exceptions.SymbolSpaceError(
                "Invalid symbol table, please ensure the ISF table produced by dwarf2json was created with version 0.8.0 or later"
            )

        known_module_addresses = cls.get_lsmod_module_addresses(
            context, vmlinux_module_name
        )
        modules_memory_boundaries = (
            linux_utilities_modules.Modules.get_modules_memory_boundaries(
                context, vmlinux_module_name
            )
        )

        yield from linux_utilities_modules.Modules.get_hidden_modules(
            context,
            vmlinux_module_name,
            known_module_addresses,
            modules_memory_boundaries,
        )

    @classmethod
    def get_hidden_modules(
        cls,
        context: interfaces.context.ContextInterface,
        vmlinux_module_name: str,
        known_module_addresses: Set[int],
        modules_memory_boundaries: Tuple,
    ) -> Iterable[interfaces.objects.ObjectInterface]:
        """Enumerate hidden modules by taking advantage of memory address alignment patterns

        This technique is much faster and uses less memory than the traditional scan method
        in Volatility2, but it doesn't work with older kernels.

        From kernels 4.2 struct module allocation are aligned to the L1 cache line size.
        In i386/amd64/arm64 this is typically 64 bytes. However, this can be changed in
        the Linux kernel configuration via CONFIG_X86_L1_CACHE_SHIFT. The alignment can
        also be obtained from the DWARF info i.e. DW_AT_alignment<64>, but dwarf2json
        doesn't support this feature yet.
        In kernels < 4.2, alignment attributes are absent in the struct module, meaning
        alignment cannot be guaranteed. Therefore, for older kernels, it's better to use
        the traditional scan technique.

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            vmlinux_module_name: The name of the kernel module on which to operate
            known_module_addresses: Set with known module addresses
            modules_memory_boundaries: Minimum and maximum address boundaries for module allocation.
        Yields:
            module objects
        """
        return linux_utilities_modules.get_hidden_modules(
            vmlinux_module_name, known_module_addresses, modules_memory_boundaries
        )

    implementation = find_hidden_modules

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Linux kernel",
                architectures=constants.architectures.LINUX_ARCHS,
            ),
            requirements.VersionRequirement(
                name="linux_utilities_modules_module_display_plugin",
                component=linux_utilities_modules.ModuleDisplayPlugin,
                version=(2, 0, 0),
            ),
            requirements.VersionRequirement(
                name="linux_utilities_modules",
                component=linux_utilities_modules.Modules,
                version=(3, 0, 1),
            ),
            requirements.BooleanRequirement(
                name="dump",
                description="Extract listed modules",
                default=False,
                optional=True,
            ),
        ]

    @staticmethod
    @deprecation.deprecated_method(
        replacement=linux_utilities_modules.Modules.get_modules_memory_boundaries,
        removal_date="2026-03-25",
        replacement_version=(3, 0, 0),
    )
    def get_modules_memory_boundaries(
        context: interfaces.context.ContextInterface,
        vmlinux_module_name: str,
    ) -> Tuple[int, int]:
        return linux_utilities_modules.Modules.get_modules_memory_boundaries(
            context, vmlinux_module_name
        )

    @deprecation.deprecated_method(
        replacement=linux_utilities_modules.Modules.get_module_address_alignment,
        removal_date="2026-03-25",
        replacement_version=(3, 0, 0),
    )
    @classmethod
    def _get_module_address_alignment(
        cls,
        context: interfaces.context.ContextInterface,
        vmlinux_module_name: str,
    ) -> int:
        """Obtain the module memory address alignment.

        struct module is aligned to the L1 cache line, which is typically 64 bytes for most
        common i386/AMD64/ARM64 configurations. In some cases, it can be 128 bytes, but this
        will still work.

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            vmlinux_module_name: The name of the kernel module on which to operate

        Returns:
            The struct module alignment
        """
        return linux_utilities_modules.get_module_address_alignment(
            context, vmlinux_module_name
        )

    @deprecation.deprecated_method(
        replacement=linux_utilities_modules.Modules.get_hidden_modules,
        removal_date="2026-03-25",
        replacement_version=(3, 0, 0),
    )
    @staticmethod
    @deprecation.deprecated_method(
        replacement=linux_utilities_modules.Modules.validate_alignment_patterns,
        removal_date="2026-03-25",
        replacement_version=(3, 0, 0),
    )
    def _validate_alignment_patterns(
        addresses: Iterable[int],
        address_alignment: int,
    ) -> bool:
        """Check if the memory addresses meet our alignments patterns

        Args:
            addresses: Iterable with the address values
            address_alignment: Number of bytes for alignment validation

        Returns:
            True if all the addresses meet the alignment
        """
        return linux_utilities_modules.validate_alignment_patterns(
            addresses, address_alignment
        )

    @classmethod
    def get_lsmod_module_addresses(
        cls,
        context: interfaces.context.ContextInterface,
        vmlinux_module_name: str,
    ) -> Set[int]:
        """Obtain a set the known module addresses from linux.lsmod plugin

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            vmlinux_module_name: The name of the kernel module on which to operate

        Returns:
            A set containing known kernel module addresses
        """
        vmlinux = context.modules[vmlinux_module_name]
        vmlinux_layer = context.layers[vmlinux.layer_name]

        known_module_addresses = {
            vmlinux_layer.canonicalize(module.vol.offset)
            for module in linux_utilities_modules.Modules.list_modules(
                context, vmlinux_module_name
            )
        }
        return known_module_addresses

    def run(self):
        return renderers.TreeGrid(
            linux_utilities_modules.ModuleDisplayPlugin.columns_results,
            self._generator(),
        )

    def _generator(self):
        yield from linux_utilities_modules.ModuleDisplayPlugin.generate_results(
            self.context,
            self.implementation,
            self.config["kernel"],
            self.config["dump"],
            self.open,
        )



# This file is Copyright 2020 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
import logging
from typing import Iterable, Optional

from volatility3.framework import renderers, interfaces, exceptions
from volatility3.framework.configuration import requirements
from volatility3.framework.interfaces import plugins
from volatility3.framework.objects import utility
from volatility3.framework.renderers import format_hints
from volatility3.framework.symbols import mac
from volatility3.plugins.mac import mount

vollog = logging.getLogger(__name__)


class List_Files(plugins.PluginInterface):
    """Lists all open file descriptors for all processes."""

    _required_framework_version = (2, 0, 0)

    @classmethod
    def get_requirements(cls):
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Kernel module for the OS",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="mount", component=mount.Mount, version=(2, 0, 0)
            ),
            requirements.VersionRequirement(
                name="mac_utilities",
                component=mac.MacUtilities,
                version=(1, 3, 0),
            ),
        ]

    @classmethod
    def _vnode_name(cls, vnode: interfaces.objects.ObjectInterface) -> Optional[str]:
        # roots of mount points have special name handling
        if vnode.v_flag & 1 == 1:
            v_name = vnode.full_path()
        else:
            try:
                v_name = utility.pointer_to_string(vnode.v_name, 255)
            except exceptions.InvalidAddressException:
                v_name = None

        return v_name

    @classmethod
    def _get_parent(cls, context, vnode):
        # root entries do not have parents
        # and parents of normal files can be smeared
        try:
            parent = vnode.v_parent.dereference()
        except exceptions.InvalidAddressException:
            return None

        if parent and not context.layers[vnode.vol.native_layer_name].is_valid(
            parent.vol.offset, parent.vol.size
        ):
            return None

        return parent

    @classmethod
    def _add_vnode(cls, context, vnode, loop_vnodes):
        """
        Adds the given vnode to loop_vnodes.

        loop_vnodes is key off the address of a vnode
        and holds its name, parent address, and object
        """

        if not context.layers[vnode.vol.native_layer_name].is_valid(
            vnode.vol.offset, vnode.vol.size
        ):
            return False

        key = vnode.vol.offset
        added = False

        if key not in loop_vnodes:
            # We can't do anything with a no-name vnode
            v_name = cls._vnode_name(vnode)
            if v_name is None:
                return added

            parent = cls._get_parent(context, vnode)
            if parent:
                parent_val = parent.vol.offset
            else:
                parent_val = None

            loop_vnodes[key] = (v_name, parent_val, vnode)

            added = True

        return added

    @classmethod
    def _walk_vnode(cls, context, vnode, loop_vnodes):
        """
        Iterates over the list of vnodes associated with the given one.
        Also traverses the parent chain for the vnode and adds each one.
        """
        added = False

        while vnode:
            if vnode in loop_vnodes:
                return added

            if not cls._add_vnode(context, vnode, loop_vnodes):
                break

            added = True

            parent = cls._get_parent(context, vnode)
            while parent and parent not in loop_vnodes:
                if not cls._walk_vnode(context, parent, loop_vnodes):
                    break

                parent = cls._get_parent(context, parent)

            try:
                vnode = vnode.v_mntvnodes.tqe_next.dereference()
            except exceptions.InvalidAddressException:
                break

        return added

    @classmethod
    def _walk_vnodelist(cls, context, list_head, loop_vnodes):
        for vnode in mac.MacUtilities.walk_tailq(list_head, "v_mntvnodes"):
            cls._walk_vnode(context, vnode, loop_vnodes)

    @classmethod
    def _walk_mounts(
        cls, context: interfaces.context.ContextInterface, kernel_module_name: str
    ) -> Iterable[interfaces.objects.ObjectInterface]:
        loop_vnodes = {}

        # iterate each vnode source from each mount
        list_mounts = mount.Mount.list_mounts(context, kernel_module_name)
        for mnt in list_mounts:
            cls._walk_vnodelist(context, mnt.mnt_vnodelist, loop_vnodes)
            cls._walk_vnodelist(context, mnt.mnt_workerqueue, loop_vnodes)
            cls._walk_vnodelist(context, mnt.mnt_newvnodes, loop_vnodes)
            cls._walk_vnode(context, mnt.mnt_vnodecovered, loop_vnodes)
            cls._walk_vnode(context, mnt.mnt_realrootvp, loop_vnodes)
            cls._walk_vnode(context, mnt.mnt_devvp, loop_vnodes)

        return loop_vnodes

    @classmethod
    def _build_path(cls, vnodes, vnode_name, parent_offset):
        path = [vnode_name]
        seen_offsets = set()

        while parent_offset in vnodes:
            parent_name, parent_offset, _ = vnodes[parent_offset]
            if parent_offset is None:
                parent_offset = 0

            # circular references from smear
            elif parent_offset in seen_offsets:
                path = []
                break

            else:
                seen_offsets.add(parent_offset)

            path.insert(0, parent_name)

        if len(path) > 1:
            path = "/".join(path)
        else:
            path = vnode_name

        if path.startswith("//"):
            path = path[1:]

        return path

    @classmethod
    def list_files(
        cls, context: interfaces.context.ContextInterface, kernel_module_name: str
    ) -> Iterable[interfaces.objects.ObjectInterface]:
        vnodes = cls._walk_mounts(context, kernel_module_name)

        for voff, (vnode_name, parent_offset, vnode) in vnodes.items():
            full_path = cls._build_path(vnodes, vnode_name, parent_offset)

            yield vnode, full_path

    def _generator(self):
        for vnode, full_path in self.list_files(self.context, self.config["kernel"]):
            yield (0, (format_hints.Hex(vnode.vol.offset), full_path))

    def run(self):
        return renderers.TreeGrid(
            [("Address", format_hints.Hex), ("File Path", str)], self._generator()
        )



# This file is Copyright 2025 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
import logging
from volatility3.framework import interfaces, deprecation
from volatility3.plugins.windows.malware import psxview

vollog = logging.getLogger(__name__)


class PsXView(
    interfaces.plugins.PluginInterface,
    deprecation.PluginRenameClass,
    replacement_class=psxview.PsXView,
    removal_date="2026-06-07",
):
    """Lists all processes found via four of the methods described in \"The Art of Memory Forensics\" which may help \
    identify processes that are trying to hide themselves.

    We recommend using -r pretty if you are looking at this plugin's output in a terminal.
    deprecated."""

    # I've omitted the desktop thread scanning method because Volatility3 doesn't appear to have the functionality
    # which the original plugin used to do it.

    # The sessions method is omitted because it begins with the list of processes found by Pslist anyway.

    # Lastly, I've omitted the pspcid method because I could not for the life of me get it to work. I saved the
    # code I do have from it, and will happily share it if anyone else wants to add it.

    _required_framework_version = (2, 0, 0)
    _version = (1, 0, 0)



# This file is Copyright 2024 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import logging
from typing import List, Iterator

from volatility3.framework import renderers, interfaces
from volatility3.framework.constants import architectures
from volatility3.framework.objects import utility
from volatility3.framework.configuration import requirements
from volatility3.framework.interfaces import plugins
from volatility3.plugins.linux import pslist

vollog = logging.getLogger(__name__)


class Ptrace(plugins.PluginInterface):
    """Enumerates ptrace's tracer and tracee tasks"""

    _required_framework_version = (2, 10, 0)
    _version = (1, 0, 2)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Linux kernel",
                architectures=architectures.LINUX_ARCHS,
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(4, 0, 0)
            ),
        ]

    @classmethod
    def enumerate_ptrace_tasks(
        cls,
        context: interfaces.context.ContextInterface,
        vmlinux_module_name: str,
    ) -> Iterator[interfaces.objects.ObjectInterface]:
        """Enumerates ptrace's tracer and tracee tasks

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            vmlinux_module_name: The name of the kernel module on which to operate

        Yields:
            A task_struct object
        """

        tasks = pslist.PsList.list_tasks(
            context,
            vmlinux_module_name,
            filter_func=pslist.PsList.create_pid_filter(),
            include_threads=True,
        )

        for task in tasks:
            if task.is_being_ptraced or task.is_ptracing:
                yield task

    def _generator(self, vmlinux_module_name):
        for task in self.enumerate_ptrace_tasks(self.context, vmlinux_module_name):
            task_comm = utility.array_to_string(task.comm)
            user_pid = task.tgid
            user_tid = task.pid
            tracer_tid = task.get_ptrace_tracer_tid() or renderers.NotAvailableValue()
            tracee_tids = task.get_ptrace_tracee_tids() or [
                renderers.NotAvailableValue()
            ]
            flags = task.get_ptrace_tracee_flags() or renderers.NotAvailableValue()

            for level, tracee_tid in enumerate(tracee_tids):
                fields = [
                    task_comm,
                    user_pid,
                    user_tid,
                    tracer_tid,
                    tracee_tid,
                    flags,
                ]
                yield (level, fields)

    def run(self):
        vmlinux_module_name = self.config["kernel"]

        headers = [
            ("Process", str),
            ("PID", int),
            ("TID", int),
            ("Tracer TID", int),
            ("Tracee TID", int),
            ("Flags", str),
        ]
        return renderers.TreeGrid(headers, self._generator(vmlinux_module_name))
